UINT32 jpeg_source_addr = 0x0000;
UINT32 jpeg_decode_addr = 0x1A000;

UINT16 picture_info[4];	// 1 = horizontal, 2 = vertical, 3 = aspect ratio

int LCD_WIDTH_1=320;
int LCD_HEIGHT_1=240;
UINT8 ori_mk_h;
UINT8 new_mk_h;

void Set_JPEG_register(void)
{

	UINT32 fsize;
	// hardcoded value : shall not be the case !
	fsize = 0xFFFFFF;	

	// clear the status of JPEG line buffer & FIFO
	SSD_REGWB(0x382, 0x07);	// Clear the status register

	// clear the JPEG control registers
	SSD_REGWB(0x386, 0x07);	// JPEG interrupt control register 0
	SSD_REGWB(0x387, 0x36);	// JPEG interrupt control register 1 -> Reflect status at 383h
	SSD_REGWB(0x3A0, 0x10);	// Trigger when more than 1/2 the FIFO size
	
	// Initi the JPEG Codec registers
	SSD_REGWB(0x402, 0x80);	// reset JPEG codec 1
	SSD_REGWB(0x400, 0x04);	// still JPEG decode, bit 5 determine MJPEG(1) or still JPEG(0)
	SSD_REGWB(0x41C, 0x02);	// error detection & revise function

	// Initi JPEG module registers
	SSD_REGWL(0x414, jpeg_source_addr/4);	// destination address -> store original JPEG data, FIFO address
	SSD_REGWB(0x3A4, 0x0F);			// 0 = 4K F = 64K for JPEG FIFO
	SSD_REGWL(0x3B8, fsize);		// set filesize
	SSD_REGWB(0x380, 0x11);			// turn on the YUV output data range 1
	//SSD_REGWB(0x380, 0x09);			// turn on the YUV output data range 1
}

void decode_jpeg_header(void)
{

	// enable JPEG Codec
	SSD_REGWB(0x380, 0x01);	// turn on the JPEG module

	// decode the header first
	Set_JPEG_register();	// set status / interrupt / flag for JPEG decode

	SSD_REGWL(0x410, jpeg_decode_addr/4);	// Start address decoded for JPEG file 32 bit addressing for FIFO engine

	// start decode process 
	SSD_REGWB(0x402, 0x01);	// write to bit 0 to start JPEG operation 1
	SSD_REGWB(0x38A, 0x00);	// check for header NOT body

	return;
	
}

char get_jpeg_file_parameters(void)
{

	UINT8 op_mode = 0;
	UINT8 temp = 0;
	UINT8 stats = 0;

	UINT16 Horizontal_Pixel = 0;
	UINT16 Vertical_Pixel = 0;
	UINT8 Operation_Status = 0;

	UINT16 H_Pixel = 0;
	UINT16 V_Pixel = 0;
	UINT32 status_quo = 0;

	UINT8 resize_x = 1;
	UINT8 resize_y = 1;
	UINT8 resize_factor = 1;

	op_mode = SSD_REGRB(0x401);		// check YUV format
	
	Operation_Status = SSD_REGRB(0x385);	// check is the decoding done
	if( Operation_Status != 0x10 )
	{
		stats = 1;	// did not decode the header
		goto end;
	}

	Vertical_Pixel = V_Pixel = SSD_REGRW(0x3DC);
	Horizontal_Pixel = H_Pixel = SSD_REGRW(0x3D8);

	if ( Horizontal_Pixel > 2560 || Vertical_Pixel > 1920 )
	{
		stats = 2;	// too big, cannot be scale down
		goto end;
	}

	// good fit	-> no resizing is needed
	if ( (310 < Horizontal_Pixel) && ( Horizontal_Pixel <= 320 ) )
	{
		if ( ( 230 < Vertical_Pixel ) && (Vertical_Pixel <= 240) )
		{
			SSD_REGWB(REG_JPEG_RESIZER_CTL,0x00);  	// REG_JPEG_RESIZER_CTL==0x360. // View Resizer Disable/Enable.		

			// remove floating window
			temp = SSD_REGRB(REG_SPECIAL_EFFECTS);	// Reg 71
			temp = temp & 0xEF;						// Bit 4 is to disable the floating window
			SSD_REGWB(REG_SPECIAL_EFFECTS, temp);
			goto end;
		}
	}	
	
	/* Consider  number of cases  */

	SSD_REGWB(REG_JPEG_RESIZER_CTL,0x01);  		// REG_JPEG_RESIZER_CTL==0x360. // View Resizer Disable/Enable.

	SSD_REGWW(REG_JPEG_RESIZER_STARTX_0,0); 	// REG_JPEG_RESIZER_STARTX_0==0x364.  // View Resizer Start X Position.
   	SSD_REGWW(REG_JPEG_RESIZER_STARTY_0,0); 	// REG_JPEG_RESIZER_STARTY_0==0x366.  // View Resizer Start Y Position.

	if ( (Horizontal_Pixel <= 309) && ( Vertical_Pixel <= 229 ))
	{
		resize_factor = 1;
		goto e_float;
	}
	
	while ( Horizontal_Pixel > LCD_WIDTH_1 )
	{
		//resize_x = Horizontal_Pixel / LCD_HEIGHT_1 	;
		resize_factor = resize_factor << 1;
		Horizontal_Pixel = Horizontal_Pixel >> 1 ;
		resize_x = resize_factor ;
	}
	resize_factor = 1;

	while ( Vertical_Pixel > LCD_HEIGHT_1 )
	{
		resize_factor = resize_factor << 1;
		Vertical_Pixel = Vertical_Pixel >> 1 ;
		resize_y = resize_factor ;
	}
	
	if( resize_x > resize_y )
	{
		resize_factor = resize_x;
	}
	else if ( resize_x < resize_y )
	{
		resize_factor = resize_y;		
	}
	else if ( resize_x == resize_y )
	{
		resize_factor = resize_x;
	}

e_float:
	SSD_REGWW(REG_JPEG_RESIZER_ENDX_0,(LCD_WIDTH_1*resize_factor)-1);  	// REG_JPEG_RESIZER_ENDX_0==0x368.
	SSD_REGWW(REG_JPEG_RESIZER_ENDY_0,(LCD_HEIGHT_1*resize_factor)-1); 	// REG_JPEG_RESIZER_ENDY_0==0x36a.
	
	SSD_REGWB(REG_JPEG_RESIZER_OP_0,resize_factor); 		// REG_JPEG_RESIZER_OP_0==0x36c.  	// Resize factor.
	SSD_REGWB(REG_JPEG_RESIZER_OP_1,2); 

	picture_info[0] = H_Pixel/resize_factor;
	picture_info[1] = V_Pixel/resize_factor;
	picture_info[2] = resize_factor;
	
	stats = 9;	// this is a flag to indicate floating window shall be shown
	
end:
	return stats;	

}

void decode_jpeg_body(void)
{
	
	// enable JPEG Codec
	SSD_REGWB(0x380, 0x01);	// turn on the JPEG module

	// decode the body now

	Set_JPEG_register();	// set status / interrupt / flag for JPEG decode

	SSD_REGWL(0x410, jpeg_decode_addr/4);	// Start address decoded for JPEG file 32 bit addressing for FIFO engine

	SSD_REGWB(0x402, 0x01);		// write to bit 0 to start JPEG operation 1
	SSD_REGWB(0x38A, 0x01);		// start the image body decode process
	return;
}

void enable_float_win(void)
{

	UINT16 Horizontal_P = picture_info[0];
	UINT16 Vertical_P 	= picture_info[1] ;
	UINT8 resize_f 		= picture_info[2];
	UINT8 check;
	char display_loc;

	UINT16 vertical_bk;
	UINT16 chk_hori_even;

	UINT8 bpp = 16;

	UINT16 x,y;

	float L_Width = LCD_WIDTH_1;
	float L_Height= LCD_HEIGHT_1;
	float display_aspect = L_Width / L_Height ;

	float aspect_ratio = 0.0;
	float hp, vp;

	check = SSD_REGRB(REG_DV_OFORMAT);	// Reg 1A4
	check = check & 0x7F;	// Bit 7 set floating window to YUV
	check = check | 0x40;	// Bit 6 set main window to RGB to create black color effect
	SSD_REGWB(REG_DV_OFORMAT, check);

	check = SSD_REGRB(REG_SPECIAL_EFFECTS);	// Reg 71
	check = check | 0x10;	// Bit 4 is to set to have floating window
	SSD_REGWB(REG_SPECIAL_EFFECTS, check);

	if( resize_f == 1)
	//if(1)
	{

		x = (L_Width - Horizontal_P)/2;
		y = (L_Height - Vertical_P)/2; 
	
		SSD_REGWW(REG_FLOAT_WIN_X_START_POS0,x/(bpp/8));
		SSD_REGWW(REG_FLOAT_WIN_X_END_POS0,((x+Horizontal_P)/(bpp/8))-1);	

		SSD_REGWW(REG_FLOAT_WIN_Y_START_POS0,y);
		SSD_REGWW(REG_FLOAT_WIN_Y_END_POS0,y+Vertical_P-1);

		SSD_REGWL(REG_FLOAT_WIN_DISP_START_ADDR0,jpeg_decode_addr/4);
				
		SSD_REGWW(REG_FLOAT_WIN_ADDR_OFFSET0,320/(32/bpp));

	}
	else
    {
		/*	In order to do scaling, the max RAM we can had is 256k-(display panel area *2) ( for 16bpp )

		 	step 1 : Check aspect ratio  ( horizontal / vertical )

   			case 1 ) if less than 1 = Tall picture
			case 2 ) if 1 = square picture
			case 3 ) if more than 1 less than 2 : rectangular picture

			step 2 : Check aspect ratio of display panel
			
			step 3 : Compare aspect ratio of display with aspect ratio of picture

					if aspect ratio of picture < aspect retio of display. 
						horizontal stretching is base on how much vertical is being stretch - case 1
					
					if aspect ratio of picture >= aspect retio of display. 
						vertical stretching is base on how much horizontal is being stretch - case 2
			
			Case 1 note : A 320 x240 display, aspect ratio = 1.33
						Thus any picture with aspect ratio 1.32 and lower, stretch to a max of 240 .

			For case 2 : Check vertical height, if 220 or more, don't need to stretch vertically further on
			         	 if 219 or less, then stretch according to how much horizontal is stretch, and maintain aspect ratio  

		*/
		hp = Horizontal_P;
		vp = Vertical_P;
		aspect_ratio = 	hp / vp;
 
		if(aspect_ratio < display_aspect )
		{
			// case 1
			display_loc = Do_resizing(Horizontal_P,Vertical_P, 1 ); 
		}
		
		if( aspect_ratio >= display_aspect )
		{
			if(Vertical_P >= 220 )	//200
			{
				// stretch horizonal only - case 2
				display_loc = Do_resizing(Horizontal_P,Vertical_P, 2 ); 
				
			}
			else
			{
				// stretch both side - case 3
				display_loc = Do_resizing(Horizontal_P,Vertical_P, 3 ); 
			}
		}
	
		
		chk_hori_even = picture_info[0] | 0x01;
		if ( chk_hori_even == picture_info[0] )
		{
			// it is odd, thus reduce by 1
			picture_info[0] = picture_info[0] - 1;
		}
		x = ( LCD_WIDTH_1 - picture_info[0] )/2;
		y = ( LCD_HEIGHT_1 - picture_info[1])/2; 

		SSD_REGWW(REG_FLOAT_WIN_X_START_POS0,x/(bpp/8));	// 0x84
		SSD_REGWW(REG_FLOAT_WIN_X_END_POS0,((x+picture_info[0])/(bpp/8))-1);	// 0x8C

		SSD_REGWW(REG_FLOAT_WIN_Y_START_POS0,y);	// 0x88
		SSD_REGWW(REG_FLOAT_WIN_Y_END_POS0,y+picture_info[1]-1);	// 0x90

		if ( display_loc == 0 )
		{
			// stretching done, display at 0
			SSD_REGWL(REG_FLOAT_WIN_DISP_START_ADDR0,0);	//0x7C
		}
		else
		{
			SSD_REGWL(REG_FLOAT_WIN_DISP_START_ADDR0, jpeg_decode_addr>>2);	// show picture at the place SD dump it too
		}

		SSD_REGWW(REG_FLOAT_WIN_ADDR_OFFSET0,320/(32/bpp));	//0x80
	}
	
   	return;
}

char Do_resizing(UINT16 Hori_P, UINT16 Vert_P, UINT8 type )
{
	UINT16 linewidth = 320;	// number of pixel
	UINT8 cnt = 4;			// this value indicate how many time the horizontal pixel is stretch; default is 4
	UINT8 ori_mk = 0;		// number of vertical pixel to make up
	UINT8 new_mk = 0;
	UINT8 expand =0;

	char display_1A = 0;	// if resize, most likely display start add at 0, but sometime no stretching, so display start add at 1A000

	float stretched_side;
	float new_ratio;
	float new_Hori_P = Hori_P;
	float new_Vert_P = Vert_P;

	UINT32 display_origin_add, pic_origin_add;

	UINT16 source_width;
	UINT16 dest_width;

	UINT16 changed_Vert_P;
	UINT16 changed_Hort_P;
	UINT16 chk_Vert_P;
	UINT16 chk_Hort_P;

	UINT16 LCD_W = LCD_WIDTH_1;
	UINT16 LCD_H = LCD_HEIGHT_1;

	UINT16 ori_quarter_Vert_P;
	UINT16 new_quarter_Vert_P;
	
	SSD_REGWB(REG_2D_SRC_WND_COLOR_MODE, 02);	// 0x1DC
	SSD_REGWB(REG_2D_DEST_WND_COLOR_MODE,02);	// 0x1DD

	switch(type)
	{
		case 1:
		{
			// stretch vertical only
			if( ((LCD_HEIGHT_1-10) < Vert_P) && ( Vert_P <= LCD_HEIGHT_1 ) )
			{
				// no need to stretch
				picture_info[0] = Hori_P;
				picture_info[1] = Vert_P;
				display_1A = 1;
				goto end;
			}

			chk_Hort_P =  Hori_P;
			chk_Hort_P = chk_Hort_P >>2;
			ori_mk_h = Hori_P - chk_Hort_P * 4;	// make up for the lost horizontal pixel

			new_ratio = LCD_HEIGHT_1/ new_Vert_P;	// get stretching factor for vertical side		
				
			stretched_side = new_ratio * Hori_P;	// use new expansion ratio, and apply to horizontal side too
			changed_Hort_P = stretched_side;

			if (changed_Hort_P < 121 )
			{
				cnt = 1;
			}
			else if (changed_Hort_P < 239 ) 
			{
				cnt = 2;
			}
			else if (changed_Hort_P > 239 ) 
			{
				cnt = 4;	
			}

			source_width = Hori_P /cnt;				// original width to stretch, must in multiply of 2

			chk_Hort_P =  changed_Hort_P;
			chk_Hort_P = changed_Hort_P >>2;
			new_mk_h = changed_Hort_P - (chk_Hort_P * 4);	// make up for the lost line

			ori_quarter_Vert_P = Vert_P >> 2;	

			new_quarter_Vert_P = LCD_HEIGHT_1 >> 2;
		
			dest_width = changed_Hort_P/cnt;		// final width to achieve
			
			picture_info[0] = changed_Hort_P;
			picture_info[1] = LCD_HEIGHT_1;

			break;
		}
		case 2:
		{
			ori_mk_h =0;
			new_mk_h =0;
			source_width = Hori_P >>2;
			dest_width = LCD_WIDTH_1/cnt;

			// stretch horizontal only
			if ( ((LCD_WIDTH_1-10) < Hori_P) && ( Hori_P <= LCD_WIDTH_1 ) )
			{
				picture_info[0] = Hori_P;
				picture_info[1] = Vert_P;
				display_1A = 1;
				goto end; // no need to stretch horizontal
			}	

			chk_Vert_P =  Vert_P;
			chk_Vert_P = chk_Vert_P >>2;
			ori_mk = new_mk = Vert_P - chk_Vert_P * 4;	// make up for the lost line

			changed_Vert_P = Vert_P;

			ori_quarter_Vert_P = Vert_P >> 2;	// since Vert_P had been divided once already

			new_quarter_Vert_P = changed_Vert_P >> 2;

			picture_info[0] = LCD_WIDTH_1;
			picture_info[1] = Vert_P;
	
			break;
		}

		case 3:
		{
			// stretch both horizontal and vertical side
			ori_mk_h =0;
			new_mk_h =0;

			source_width = Hori_P >>2;
			dest_width = LCD_WIDTH_1/cnt;

			new_ratio = LCD_WIDTH_1/ new_Hori_P;	// get stretching factor for vertical side		
				
			stretched_side = new_ratio * Vert_P;	// use new horizontal expansion ratio, and apply to vertical side too
			changed_Vert_P = stretched_side;

			chk_Vert_P =  Vert_P;
			chk_Vert_P = chk_Vert_P >>2;
			ori_mk = Vert_P - chk_Vert_P * 4;	// make up for the lost line				

			ori_quarter_Vert_P = Vert_P >> 2;	// since Vert_P had been divided once already

			chk_Vert_P =  changed_Vert_P;
			chk_Vert_P = chk_Vert_P >>2;
			new_mk = changed_Vert_P - (chk_Vert_P * 4);	// make up for the lost line

			new_quarter_Vert_P = changed_Vert_P >> 2;

			picture_info[0] = LCD_WIDTH_1;
			picture_info[1] = changed_Vert_P;

			break;
		}

		default:
				break;
	}

	/* set up address for various copy and stretching */
	
	pic_origin_add = jpeg_decode_addr >> 1;		// stretching add format != frame display add format
												// this is the origin add of the picture
	
	display_origin_add = 0;		// this is the start add for display on LCD					
	
	while (expand < 3)
	{
		stretch (ori_quarter_Vert_P, source_width, new_quarter_Vert_P, dest_width, linewidth, 
				pic_origin_add+(LCD_WIDTH_1*ori_quarter_Vert_P*expand),
	 			display_origin_add + (LCD_WIDTH_1 * new_quarter_Vert_P*expand), cnt );	
		expand++;
	}
	// this is to make up for those vertical line not stretched.
	stretch (ori_quarter_Vert_P+ori_mk, source_width, new_quarter_Vert_P+new_mk, dest_width, linewidth, 
				pic_origin_add+(LCD_WIDTH_1*ori_quarter_Vert_P*expand),
	 			display_origin_add + (LCD_WIDTH_1 * new_quarter_Vert_P*expand), cnt );	

end:
	return display_1A;
}

void Off_jpeg(void)
{
	delay(50);
	// stop the JPEG module
	SSD_REGWB(0x380,0x10);		// the JPEG module is disabled and the clock source is disabled.
    SSD_REGWB(0x402,0x00);   	// No JPEG SW reset, No JPEG Start Operation.
    SSD_REGWB(0x38A,0x00); 		// YUV data capturing stops at the end of the current frame.
	SSD_REGWB(REG_JPEG_RESIZER_CTL,0x00); 	// Off the resizer
	
	// in case this values get corrupted, after JPEG is off, we reset it again.
	LCD_WIDTH_1  =320;
	LCD_HEIGHT_1 =240;

	return;
}

void stretch (UINT16 src_height, UINT16 src_length, UINT16 dst_height, UINT16 dst_length, UINT16 pixel_in_1, 
			  UINT32 source_add, UINT32 dest_add, UINT8 round)
{

	UINT8 rnd = 0;
	UINT8 mk_h;
	UINT8 new_h;

	UINT16 chk_src_len_even;
	UINT16 chk_dst_len_even;

	UINT16 final_src_length;
	UINT16 final_dst_length;

	// 2D engine limitation, YUV picture, source and destination length must be even
	chk_src_len_even = src_length;
	chk_src_len_even = chk_src_len_even | 0x1;	// check is it odd
	if ( chk_src_len_even == src_length )	// it is odd
	{
		src_length = src_length + 1 ;
		// will make sure final src_length stretch is also even number
		if ( ori_mk_h == 1 )
		{
			mk_h = ori_mk_h + (round-1);			
		}
		else if ( ori_mk_h == 0 || ori_mk_h == 2 ) 
		{
			mk_h = round;
		}
		else if ( ori_mk_h == 3)
		{
			mk_h = ori_mk_h - 1;
		}
		final_src_length = src_length - mk_h;
	}
	else
	{
		// it is even, take care of ori_mk_h problem
		if ( ori_mk_h <= 1 )	
		{
			mk_h = 0;
		}
		else if ( ori_mk_h >= 3 )
		{
			mk_h = 2;
		}
		final_src_length = src_length + mk_h;
	}

	// destination side
	chk_dst_len_even = dst_length;
	chk_dst_len_even = chk_dst_len_even | 0x1;	// check is it odd
	if ( chk_dst_len_even == dst_length )	// it is odd
	{
		dst_length = dst_length + 1 ;
		// will make sure final src_length stretch is also even number
		if ( new_mk_h == 1 )
		{
			new_h = new_mk_h + (round-1);			
		}
		else if ( new_mk_h == 0 || new_mk_h == 2 ) 
		{
			new_h = round;
		}
		else if ( new_mk_h == 3)
		{
			new_h = new_mk_h - 1;
		}
		final_dst_length = dst_length - new_h;
	}	
	else	// it is even
	{
		// it is even, take care of new_mk_h problem
		if ( new_mk_h <= 1 )	
		{
			new_h = 0;
		}
		else if ( new_mk_h >= 3 )
		{
			new_h = 2;
		}
		final_dst_length = dst_length + new_h;
	}

	
	SSD_REGWB(REG_2D_CMD_2,0x08);	// block stretch

	while ( rnd < (round-1) )
	{
		// do the resizing by 1/2 picture first
		SSD_REGWW(REG_2D_SRC_WND_HEIGHT0,src_height);	  		// 0x1E8
		SSD_REGWW(REG_2D_DEST_WND_HEIGHT0,dst_height);	  	// 0x1F0

		SSD_REGWW(REG_2D_SRC_WND_ADDR_OFFSET0,pixel_in_1);   // 0x1D8
		SSD_REGWW(REG_2D_DEST_WND_ADDR_OFFSET0,pixel_in_1);  // 0x1F8

		SSD_REGWW(REG_2D_SRC_WND_WIDTH0,src_length);  //   0x1E4  width in term of pixel
		SSD_REGWW(REG_2D_DEST_WND_WIDTH0,dst_length); //  	0x1EC

		SSD_REGWL(REG_2D_SRC_WND_START_ADDR0 ,(source_add+(rnd*src_length)));		// 0x1D4
		SSD_REGWL(REG_2D_DEST_WND_START_ADDR0,(dest_add+(rnd*dst_length)));	// 0x1F4

		SSD_REGWB(REG_2D_CMD_FIFO_STATUS,0x81);

		delay(5);	  
		rnd++;	
	}
	// the last stretch is to compensate for any miss out horizontal pixel due to dividing an odd pixel
	SSD_REGWW(REG_2D_SRC_WND_HEIGHT0,src_height);	  		// 0x1E8
	SSD_REGWW(REG_2D_DEST_WND_HEIGHT0,dst_height);	  	// 0x1F0

	SSD_REGWW(REG_2D_SRC_WND_ADDR_OFFSET0,pixel_in_1);   // 0x1D8
	SSD_REGWW(REG_2D_DEST_WND_ADDR_OFFSET0,pixel_in_1);  // 0x1F8

	SSD_REGWW(REG_2D_SRC_WND_WIDTH0,final_src_length);  //   0x1E4  width in term of pixel
	SSD_REGWW(REG_2D_DEST_WND_WIDTH0,final_dst_length); //  	0x1EC

	SSD_REGWL(REG_2D_SRC_WND_START_ADDR0 ,(source_add+(rnd*src_length)));		// 0x1D4
	SSD_REGWL(REG_2D_DEST_WND_START_ADDR0,(dest_add+(rnd*dst_length)));	// 0x1F4

	SSD_REGWB(REG_2D_CMD_FIFO_STATUS,0x81);

	delay(5);	  
	return;
}

